/* 
   TI-86 String Routines
   by David Phillips <david@acz.org>
   http://www.acz.org/
 
   When you call these routines, the name should be a 1-8 byte string
   containing appropriate characters for a variable or filename.
   '.86s' will be appended to it for the filename.  This will be the 
   name used for the variable when on the calculator.
  
   To avoid memory allocation problems/hassles, a 16k buffer, str_read,
   is used to read in a string.  You can either use this directly, or
   copy it to another location.  Then length of the string is returned.

   Examples:
      write_string("map", map, 256);

      len = read_string("map");
      memcpy(map, str_read, len);
*/

#include <stdio.h>
#include <stdlib.h>

char str_read[16384];

//======== TYPEDEFS/STRUCTS ===============================================
typedef struct
{
  char  sig[8],         // "**TI86**"
        ext_sig[3],     // EOF, CR, NULL byte (1a, 0a, 00)
        comment[42];    // padded with NULL bytes to len of 42
  short file_len;       // file length minus 0x39
} FileHeader;

typedef struct
{
  short data_ptr,       // points to var_len2 at end of entry
        var_len;        // length of variable
  char  type,           // variable type
        name_len,       // variable name length
        name[8];        // variable name
  short var_len2;       // exactly the same as var_len
} FileEntry;

typedef struct
{
  short len;            // length of the program
  char  token[2];       // for a compiled asm prgm, (8e, 28)
} VarPrgm;

typedef struct
{
  short len;            // length of the string
} VarString;

//======== STRING FUNCTIONS ===============================================
void write_string(char *name, char *data, short len)
{
  FILE  *fp;
  short x;
  unsigned short checksum;
  FileHeader header;
  char fname[16],
       vname[9];
  const char HEADER[] = {"Created with TI-86 String Routines"};

  strcpy(fname, name);
  strcat(fname, ".86s");
  
  memcpy(header.sig, "**TI86**", 8);
  header.ext_sig[0] = 0x1a;
  header.ext_sig[1] = 0x0a;
  header.ext_sig[2] = 0x00;
  memset(header.comment, 0, 42);
  memcpy(header.comment, HEADER, strlen(HEADER));
  header.file_len = sizeof(FileEntry) + len + 2;

  fp = fopen(fname, "wb");
  if (fp == NULL)
    printf("\nError opening %s.  String not saved!\n", fname);
  else
  {
    fwrite(&header, sizeof(header), 1, fp);

    checksum = 0;
    fputc_sum(12, fp, &checksum);
    fputc_sum(0, fp, &checksum);
    fputc_sum((len + 2) % 256, fp, &checksum);
    fputc_sum((len + 2) >> 8, fp, &checksum);
    fputc_sum(0x0c, fp, &checksum);
    fputc_sum(strlen(name), fp, &checksum);
    memset(vname, 0, 9);    
    memcpy(vname, name, strlen(name));
    for(x = 0; x < 8; x++)
      fputc_sum(vname[x], fp, &checksum);
    fputc_sum((len + 2) % 256, fp, &checksum);
    fputc_sum((len + 2) >> 8, fp, &checksum);
    fputc_sum(len % 256, fp, &checksum);
    fputc_sum(len >> 8, fp, &checksum);

    for(x = 0; x < len; x++)
      fputc_sum(data[x], fp, &checksum);
    fwrite(&checksum, 2, 1, fp);

    fclose(fp);
    if (errno != 0)
    {
      printf("\nError writting %s.  String not saved!\n", fname);
      return;
    }
    printf("String saved as %s\n", fname);
  }
}

fputc_sum(char c, FILE *fp, unsigned short *sum)
{
  *sum += c;
  fputc(c, fp);
}

// returns length
short read_string(char *name)
{
  FILE        *fp;
  FileHeader  header;
  FileEntry   entry;
  VarString   str;
  char        *fname,
              *data;

  strcpy(fname, name);
  strcat(fname, ".86s");

  fp = fopen(fname, "rb");
  if (fp == NULL)
  {
    printf("\nError opening %s.  String not loaded!\n", fname);
    return NULL;
  }
  else
  {
    fread(&header, sizeof(FileHeader), 1, fp);
    fread(&entry, sizeof(FileEntry), 1, fp);
    fread(&str, sizeof(VarString), 1, fp);

    fread(&str_read, str.len, 1, fp);

    fclose(fp);
    if (errno != 0)
    {
      printf("\nError reading %s.  String not loaded!\n", fname);
      return NULL;
    }
    return str.len;
  }
}
